# !/usr/bin/env python
# -*-coding:utf-8 -*-
# Author     ：Campanula 梦芸 何
import abc
import os
import threading
import time
from typing import List, Dict

from anduin.common import dbg, get_obj_name
from anduin.frames.client_base import ClientBase

TIMEOUT = 30

class ManagerBase(abc.ABC):
    def __init__(self, db_config):
        self.use_cache = False
        self.t_data = db_config
        self.host = db_config.get('host')
        self.user = db_config.get('user')
        self.password = db_config.get('password')
        self.database = db_config.get('database')
        self.charset = db_config.get('charset')
        self.port = db_config.get('port')
        self.connect_pool = {}
        # {
        #     tid:{
        #         connect_sid:connect
        #     }
        # }
        self.out_time_pool = {}
        # {
        #     tid:[connect_sid]
        # }
        # dbg('creating DB connection pool...',get_obj_name(self))

    @staticmethod
    def get_cur_thread_id() -> str:
        mypid = str(os.getpid())
        mytid = str(threading.currentThread().ident)
        thread_id = mypid + '.' + mytid
        return thread_id

    def get_cur_client_pool_by_thread_id(self) -> Dict[int, ClientBase]:
        tid = ManagerBase.get_cur_thread_id()
        if tid not in self.connect_pool:
            self.connect_pool[tid] = {}
        return self.connect_pool.get(tid)

    def get_cur_outtime_pool_by_thread_id(self) -> List[int]:
        tid = ManagerBase.get_cur_thread_id()
        if tid not in self.out_time_pool:
            self.out_time_pool[tid] = []
        return self.out_time_pool.get(tid)

    @abc.abstractmethod
    def create_connection(self):
        pass

    @abc.abstractmethod
    def get_free_client(self):
        pass

    def clean_pool(self):
        cur_time = int(time.time())
        cur_out_time_pool = self.get_cur_outtime_pool_by_thread_id()
        cur_pool = self.get_cur_client_pool_by_thread_id()
        print(cur_out_time_pool)
        print(cur_pool)
        for sid,client in cur_pool.items():
            if cur_time - client.last_connect_time > TIMEOUT:
                if client.is_lock is False:
                    cur_out_time_pool.append(sid)
        for sid in cur_out_time_pool:
            if sid in cur_pool:
                cur_pool.pop(sid)






# !/usr/bin/env python
# -*-coding:utf-8 -*-
# Author     ：Campanula 梦芸 何
import asyncio
import time


from anduin.common import dbg,get_db_index,ENGINE_MYSQL
from anduin.db.sql.async_mysql.db_client import AsyncMySQLClient
from anduin.frames.manager_base import ManagerBase, TIMEOUT


class AsyncMySQLManager(ManagerBase):
    def __init__(self, config):
        super().__init__(config)

    async def create_connection(self):
        dbg('无可用空闲链接，创建链接...', get_db_index(self.t_data))
        db_client = AsyncMySQLClient(self.host, self.user, self.password, self.port, self.database, ENGINE_MYSQL,
                                     self.charset)
        await db_client.connect_db()
        my_pool = self.get_cur_client_pool_by_thread_id()
        sid = id(db_client)
        my_pool[sid] = db_client
        return db_client

    async def get_free_client(self):
        self.clean_pool()
        my_client = None
        cur_time = int(time.time())
        my_pool = self.get_cur_client_pool_by_thread_id()
        for sid, client in my_pool.items():
            if cur_time - client.last_connect_time < TIMEOUT:
                if client.is_lock is False:
                    my_client = client
                    break
        if my_client is None:
            my_client = await self.create_connection()
        my_client.lock()
        return my_client




import datetime
import functools
import inspect
import os
import random
import sqlite3
import string
import sys
import time
from functools import wraps
from typing import Any

import aredis
import pymysql

ENGINE_SQLITE = 'sqlite'
ENGINE_MYSQL = 'mysql'
ENGINE_REDIS = 'async_redis'
danger_sig = ['--', '-+', '#']
start_time = int(time.time())
ENGINE_DICT = {
    ENGINE_MYSQL: pymysql,
    ENGINE_SQLITE: sqlite3,
    ENGINE_REDIS:aredis,
}
sql_clean_time = 45
LOG_PATH = '%s/.anduin/' % os.path.expanduser('~')


def time_to_str(times):
    date_array = datetime.datetime.utcfromtimestamp(times + (8 * 3600))
    return date_array.strftime("%Y-%m-%d %H:%M:%S")


def get_filename():
    if sys.platform != 'win32':
        fn = '%s%s-%s.log' % (LOG_PATH, sys.argv[0].split('/')[-1], start_time)
    else:
        fn = '%s\\.anduin\\%s-%s.log' % (os.path.expanduser('~'), sys.argv[0].split('\\')[-1], start_time)
    return fn

fn = get_filename()
if sys.platform != 'win32':
    try:
        os.mkdir('%s/.anduin' % os.path.expanduser('~'))
    except Exception as e:
        print(str(e))
    print('系统内核为',sys.platform,'本插件调用日志保存在%s' % get_filename())
else:
    print('windows下不支持持久化日志，请自行控制台生产')
fh = open(fn, 'a')


if sys.platform != 'win32':
    write_log = True
else:
    write_log = False


def dbg(*args):
    res = ['[%s Anduin Engine]' % time_to_str(int(time.time()))] + list(args)
    print(*res)
    if write_log:
        for i in res:
            fh.write(str(i) + ' ')
        fh.write('\n')
        # fh.close()


def get_db_index(db_config_dict):
    return '%s@%s:%s'%(db_config_dict.get('user'),db_config_dict.get('host'),db_config_dict.get('database'))


def func_time(f):
    """
    简单记录执行时间
    :param f:
    :return:
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = f(*args, **kwargs)
        end = time.time()
        info = str(f.__name__) + ' took ' + str(end - start) + ' seconds '
        dbg(info)
        return result

    return wrapper


def can_return_directly(res):
    if res is None:
        return True
    if res is False:
        return True
    if isinstance(res, Exception):
        return True


def clean_old_log(filename='*.log'):
    if sys.platform != 'win32':
        os.system('rm %s%s' % (LOG_PATH, filename))


def async_decorators(method):
    """
    异步装饰器装饰异步方法
    :param method: 被装饰协程（异步方法）
    :return:
    """

    @functools.wraps(method)
    async def wrapper(*args, **kwargs):
        # 异步执行
        # 此处必须await 切换调用被装饰协程，否则不会运行该协程
        data = method(*args, **kwargs)
        return data

    return wrapper

def get_obj_name(obj: Any) -> str:
    """获取对象的名称"""
    if inspect.isfunction(obj) or inspect.isclass(obj):
        return obj.__name__
    return obj.__class__.__name__

def create_rand_string(length=16):
    ran_str = ''.join(random.sample(string.ascii_letters + string.digits, length))
    return ran_str


# !/usr/bin/env python
# -*-coding:utf-8 -*-
# Author     ：Campanula 梦芸 何
from typing import Iterable

import aiomysql
from aiomysql import DictCursor


from anduin.common import dbg, get_obj_name
from anduin.common import ENGINE_MYSQL
from anduin.parser.sql_parser import Parser
from anduin.frames.client_base import ClientBase


# from pymysql.cursors import DictCursor


class AsyncMySQLClient(ClientBase):
    def __init__(self, *args):
        super().__init__(*args)
        self.db = None
        self._tables = {}

    async def connect_db(self):
        try:
            res = await aiomysql.connect(host=self._host, user=self._user, password=self._psw, db=self._dbname,
                                     charset=self._charset, port=self._port)
        # dbg('连接成功',type(res))
            dbg('连接创建成功', get_obj_name(self))
            self.db = res
            return res
        except Exception as e:
            dbg('连接创建失败',e,get_obj_name(self))
            return


    async def load_an_table(self, tablename):
        sql = 'show fields from ' + tablename
        res = await self.query(sql, show_sql=False)
        if isinstance(res,Iterable):
            self._tables[tablename] = list(map(lambda x: x[0],res))
        return res

    async def _load_tables(self, show_sql=False):
        sql = 'show tables'
        res = await self.query(sql, show_sql)
        tables = list(map(lambda x: x[0], res))
        for table in tables:
            table_name = self._dbname + '.' + table
            await self.load_an_table(table_name)

    async def release_lock(self):
        await self.commit()
        self.is_lock = False

    async def commit(self):
        await self.db.commit()

    async def query(self, sql: str, show_sql=None, sql_params=None, return_dict=False):
        dummy_sql = sql % tuple(sql_params) if sql_params is not None else sql
        if return_dict is True:
            cursor = await self.db.cursor(DictCursor)
        else:
            cursor = await self.db.cursor()
        if show_sql is True:
            dbg('sql_id', id(self), dummy_sql)
        try:
            if sql_params is not None:
                await cursor.execute(sql,tuple(sql_params))
            else:
                await cursor.execute(sql)
            self.update_last_execute_time()
            results = await cursor.fetchall()
        except Exception as e:
            dbg('<--------DBERROR-------->')
            dbg(dummy_sql)
            dbg('execute fail!', str(e))
            dbg('<--------DBERROR-------->')
            results = e
        return results

    async def create(self, table, columns, table_comment='', show_sql=False):
        sql = Parser.create_table_parser(table, columns, table_comment, sql_engine=ENGINE_MYSQL)
        await self.query(sql, show_sql)
        # asyncio.get_event_loop()
        await self.commit()
        return

    async def find(self, table, conditions, or_cond=None, fields=('*',), order=None, show_sql=False, for_update=False):
        if table not in self._tables:
            await self.load_an_table(table)

        if fields[0] == '*' and len(fields) == 1:
            fieldList = self._tables[table]
            fields = fieldList

        sql, sql_params = Parser.find_info(table, conditions, or_cond, fields, None, order, None, for_update)
        if sql is None:
            return

        sql += " limit 1"
        # self.db.commit()
        res = await self.query(sql, show_sql, sql_params, return_dict=True)
        if isinstance(res, Exception) is True:
            return res

        if 0 == len(res):
            return None

        result = res[0]
        return result

    # 查找数据
    async def select(self, table, conditions, or_cond=None, fields=('*',), group=None, order=None, limit=None,
                     show_sql=False,
                     for_update=False):
        if table not in self._tables:
            await self.load_an_table(table)

        sql, sql_params = Parser.find_info(table, conditions, or_cond, fields, group, order, limit, for_update)
        if sql is None:
            return
        #
        res = await self.query(sql, show_sql, sql_params, return_dict=True)
        if isinstance(res, Exception) is True:
            return res

        if 0 == len(res):
            return None

        return res

    async def insert(self, table, content, show_sql=False):
        sql, sql_params = Parser.insert_parser(table, content)
        r = await self.query(sql, show_sql, sql_params)
        return r

    async def update(self, table, conditions, or_cond=None, params=None, show_sql=False):
        # dbg('开始执行')
        sql, sql_params = Parser.update_parser(table, conditions, or_cond, params)
        r = await self.query(sql, show_sql, sql_params)
        # dbg('自动提交完毕')
        return r

    async def delete(self, table, condition, or_cond=None, show_sql=False):
        sql = 'delete from %s where  ' % table
        sql, sql_params = Parser.bind_conditions(sql, condition, or_cond)
        #  #
        r = await self.query(sql, show_sql, sql_params)
        return r



# !/usr/bin/env python
# -*-coding:utf-8 -*-
# Author     ：Campanula 梦芸 何
import abc
import time


class ClientBase(abc.ABC):
    def __init__(self, host='127.0.0.1', user='', psw='', port=0, dbname='', engine='mysql', charset='utf8mb4'):
        self.id = id(self)
        self._host = host
        self._user = user
        self._psw = psw
        self._dbname = dbname
        self._port = port
        self._engine = engine
        self._charset = charset
        self.db_engine = engine
        self.is_lock = False
        self.last_connect_time = int(time.time())
        self.db = None

    @abc.abstractmethod
    def connect_db(self):
        pass

    def lock(self):
        self.is_lock = True

    def release_lock(self):
        self.commit()
        self.is_lock = False

    def commit(self):
        self.db.commit()

    def update_last_execute_time(self):
        self.last_connect_time = int(time.time())



# !/usr/bin/env python
# -*-coding:utf-8 -*-
# Author     ：Campanula 梦芸 何
import time

from anduin.common import ENGINE_REDIS,dbg,get_db_index
from anduin.db.no_sql.async_redis.redis_client import AsyncRedisClient
from anduin.frames.manager_base import ManagerBase, TIMEOUT


class AsyncRedisManager(ManagerBase):
    def __init__(self,config):
        super().__init__(config)

    def create_connection(self)->AsyncRedisClient:
        dbg('无可用空闲链接，创建链接...', get_db_index(self.t_data))
        db_client = AsyncRedisClient(self.host, self.user, self.password, self.port, self.database, ENGINE_REDIS,
                                self.charset)
        my_pool = self.get_cur_client_pool_by_thread_id()
        sid = id(db_client)
        my_pool[sid] = db_client
        return db_client

    def get_free_client(self)->AsyncRedisClient:
        self.clean_pool()
        my_client = None
        cur_time = int(time.time())
        my_pool = self.get_cur_client_pool_by_thread_id()
        for sid, client in my_pool.items():
            if cur_time - client.last_connect_time < TIMEOUT:
                if client.is_lock is False:
                    my_client = client
                    break
        if my_client is None:
            my_client = self.create_connection()
        my_client.lock()
        return my_client






# !/usr/bin/env python
# -*-coding:utf-8 -*-
# Author     ：Campanula 梦芸 何
from anduin.common.tools import ENGINE_SQLITE, ENGINE_MYSQL


class Parser(object):
    @staticmethod
    def create_table_parser(table, columns, table_comment='', sql_engine=ENGINE_MYSQL):
        sql = 'create table %s(' % table
        tail = ''
        for item in columns:
            col = ''
            if sql_engine == ENGINE_SQLITE:
                item = item[:-1]
            for i in item:
                col += str(i)
                col += ' '
            col += ','
            tail += col

        tail = tail[:-1] + ')'
        sql += tail
        sql += 'charset=utf8mb4,engine=innodb,comment="%s"' % table_comment
        if sql_engine == ENGINE_MYSQL:
            sql += 'charset=utf8mb4,engine=innodb,comment="%s"' % table_comment
        if sql_engine == ENGINE_SQLITE:
            sql = sql.replace('int AUTO_INCREMENT primary key', 'INTEGER PRIMARY KEY AUTOINCREMENT')
        return sql

    # 拼接create sql

    @staticmethod
    def insert_parser(table, content):
        params = content
        keys = str(tuple(params.keys()))
        keys = keys.replace("'", "")
        if 1 == len(params):
            keys = keys[:-2] + ")"
        sql_params = []
        sql = 'insert into %s%s values (' % (table, keys)
        for i in params.values():
            sql_params.append(i)
        for _ in params.values():
            sql += '%s,'
        sql = sql[:-1]
        sql += ')'
        return sql, sql_params

    # insert规则拼接

    @staticmethod
    def find_info(table, conditions, or_cond, fields=None, group=None, order=None, limit=None, for_update=False):

        fields = list(fields)
        # print(fields)
        sql = 'select %s from %s where  ' % (','.join(fields), table)
        # dbg(sql)
        sql, sql_params = Parser.bind_conditions(sql, conditions, or_cond)

        if group is not None:
            sql += 'group by '
            for i in group:
                sql += '%s ,' % i
            sql = sql[:-1]

        if order is not None:
            sql += 'order by '
            count = 0
            for i in order:
                if count % 2 == 0:
                    sql += '%s ' % i
                else:
                    sql += '%s ,' % i
                count += 1
            sql = sql[:-1]

        if limit is not None:
            sql += 'limit '
            for i in limit:
                sql += '%s, ' % i
            sql = sql[:-2]

        if for_update is True:
            sql += ' for update'
        return sql, sql_params

    # select 规则拼接

    @staticmethod
    def bind_conditions(sql, conditions, or_cond):
        sql_params = []
        if conditions is None:
            conditions = []
        if or_cond is None:
            or_cond = []
        if len(or_cond) + len(conditions) == 0:
            return sql[:-7], None

        if len(conditions) > 0:
            for unit in conditions:
                value = unit[2]
                sql = sql + " %s %s binary " % (unit[0], unit[1]) + '%s'
                sql_params += [value]
                sql += "  and "
                if 'in' in unit[1]:
                    sql = sql.replace('binary', '')
            sql = sql[:-4]
            if len(or_cond) > 0:
                sql += ' or '

        if len(or_cond) > 0:
            for unit in or_cond:
                value = unit[2]
                sql = sql + " %s %s binary " % (unit[0], unit[1]) + '%s'
                sql_params += [value]
                sql += "  or "
                if 'in' in unit[1]:
                    sql = sql.replace('binary', '')
            sql = sql[:-3]

        return sql, sql_params

    @staticmethod
    def update_parser(table, conditions, or_cond, params):
        if params == {} or params is None:
            # dbg('没有params，结束执行')
            return
        sql = 'update %s set ' % table
        sql_params_header = []
        for param, value in params.items():
            sql = sql + " %s = " % param + '%s,'
            sql_params_header += [value]

        sql = sql[:-1] + ' where  '
        sql, sql_params = Parser.bind_conditions(sql, conditions, or_cond)
        # dbg(sql)
        if sql_params is None:
            sql_params = []
        sql_params = sql_params_header + sql_params
        return sql, sql_params





import codecs
import os
from setuptools import setup, find_packages

VER = "8.3.3"

def read(fname):
    return codecs.open(os.path.join(os.path.dirname(__file__), fname),encoding='utf-8').read()

if __name__ == '__main__':
    setup(
        # 以下为必需参数
        name='anduin',  # 模块名
        version=VER,  # 当前版本
        description='a lite mysql & sqlite3 & async_redis connect engine, mapping table into k-v structure , support async work',  # 简短描述
        license='MIT',
        long_description=read("README.md"),
        author='campanula',
        author_email='campanulamediuml@gmail.com',
        platforms='any',
        keywords="mysql , sqlite3 , sql engine , orm , async_redis",
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
        ],
        url='https://github.com/campanulamediuml/Anduin',
        install_requires=[
            'pymysql>=0.9',
            'aredis>=1.1.0',
            'aiomysql>=0.0.21',
            'redis>=4.0.0'
        ],
        include_package_data=True,
        zip_safe=True,
        packages=find_packages(),
        python_requires='>=3.6',
    )



# !/usr/bin/env python
# -*-coding:utf-8 -*-
# Author     ：Campanula 梦芸 何
import json
import time
from typing import Union, Dict

import aredis
from aredis import StrictRedis

from anduin.frames.client_base import ClientBase
from anduin.common import dbg,get_obj_name


class AsyncRedisClient(ClientBase):
    def __init__(self,*args):
        super().__init__(*args)
        self.db:Union[StrictRedis,None] = self.connect_db()

    def lock(self):
        self.is_lock = True

    def unlock(self):
        self.is_lock = False

    def connect_db(self)->Union[StrictRedis,None]:
        try:
            res = aredis.StrictRedis(host=self._host, port=self._port, password=self._psw,db=self._dbname,)
            dbg('连接创建成功', get_obj_name(self))
            return res
        except Exception as e:
            dbg('连接创建失败', e, get_obj_name(self))
            return


    async def hmget(self,key:str,*args) -> Dict[str,Dict]:
        res = await self.db.hmget(key,*args)
        r = {}
        for i in range(0, len(args)):
            k = args[i]
            v = res[i - 1]
            # print(k, v, v.decode('utf-8'))
            r[k] = json.loads(v.decode('utf-8').replace("'", '"'))
        self.last_connect_time = int(time.time())
        return r

    async def hmset(self,*args):
        res = await self.db.hmset(*args)
        self.last_connect_time = int(time.time())
        return res
    async def get(self,*args):
        res = await self.db.get(*args)
        if isinstance(res,bytes):
            return res.decode('utf-8')
        self.last_connect_time = int(time.time())
        return res

    async def set(self,*args):
        res = await self.db.set(*args)
        self.last_connect_time = int(time.time())
        return res

    async def expire(self,*args):
        res = await self.db.expire(*args)
        self.last_connect_time = int(time.time())
        return res







